@@ -1,6 +1,6 @@
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2013 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2014 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -74,42 +74,31 @@ static const char *mproc =
 #endif
 	;
 
-#ifdef INET6
-static char **restore;
-static ssize_t nrestore;
-#endif
-
-char *
-hardware_platform(void)
+int
+hardware_platform(char *str, size_t len)
 {
 	FILE *fp;
-	char *buf, *p;
+	char buf[256];
 
 	if (mproc == NULL) {
 		errno = EINVAL;
-		return NULL;
+		return -1;
 	}
 
 	fp = fopen("/proc/cpuinfo", "r");
 	if (fp == NULL)
-		return NULL;
+		return -1;
 
-	p = NULL;
-	while ((buf = get_line(fp))) {
+	while (fscanf(fp, "%255s : ", buf) != EOF) {
 		if (strncmp(buf, mproc, strlen(mproc)) == 0) {
-			p = strchr(buf, ':');
-			if (p != NULL && ++p != NULL) {
-				while (*p == ' ')
-					p++;
-				break;
-			}
+			fscanf(fp, "%255s", buf);
+			fclose(fp);
+			return snprintf(str, len, ":%s", buf);
 		}
 	}
 	fclose(fp);
-
-	if (p == NULL)
-		errno = ESRCH;
-	return p;
+	errno = ESRCH;
+	return -1;
 }
 
 #ifdef INET6
@@ -117,16 +106,15 @@ static int
 check_proc_int(const char *path)
 {
 	FILE *fp;
-	char *buf;
+	int i;
 
 	fp = fopen(path, "r");
 	if (fp == NULL)
 		return -1;
-	buf = get_line(fp);
+	if (fscanf(fp, "%d", &i) != 1)
+		i = -1;
 	fclose(fp);
-	if (buf == NULL)
-		return -1;
-	return atoi(buf);
+	return i;
 }
 
 static ssize_t
@@ -145,51 +133,34 @@ write_path(const char *path, const char *val)
 
 static const char *prefix = "/proc/sys/net/ipv6/conf";
 
-static void
-restore_kernel_ra(void)
+void
+restore_kernel_ra(struct dhcpcd_ctx *ctx)
 {
 	char path[256];
 
-#ifndef DEBUG_MEMORY
-	if (options & DHCPCD_FORKED)
-		return;
-#endif
-
-	for (nrestore--; nrestore >= 0; nrestore--) {
-#ifdef DEBUG_MEMORY
-		if (!(options & DHCPCD_FORKED)) {
-#endif
-		syslog(LOG_INFO, "%s: restoring Kernel IPv6 RA support",
-		       restore[nrestore]);
-		snprintf(path, sizeof(path), "%s/%s/accept_ra",
-			 prefix, restore[nrestore]);
-		if (write_path(path, "1") == -1 && errno != ENOENT)
-			syslog(LOG_ERR, "write_path: %s: %m", path);
-#ifdef DEBUG_MEMORY
+	for (; ctx->ra_restore_len > 0; ctx->ra_restore_len--) {
+		if (!(ctx->options & DHCPCD_FORKED)) {
+			syslog(LOG_INFO, "%s: restoring Kernel IPv6 RA support",
+			    ctx->ra_restore[ctx->ra_restore_len - 1]);
+			snprintf(path, sizeof(path), "%s/%s/accept_ra",
+			    prefix, ctx->ra_restore[ctx->ra_restore_len - 1]);
+			if (write_path(path, "1") == -1 && errno != ENOENT)
+			    syslog(LOG_ERR, "write_path: %s: %m", path);
 		}
-		free(restore[nrestore]);
-#endif
+		free(ctx->ra_restore[ctx->ra_restore_len - 1]);
 	}
-#ifdef DEBUG_MEMORY
-	free(restore);
-#endif
+	free(ctx->ra_restore);
+	ctx->ra_restore = NULL;
 }
 
 int
-check_ipv6(const char *ifname, int own)
+check_ipv6(struct dhcpcd_ctx *ctx, const char *ifname, int own)
 {
-	static int ipv6_checked = 0;
-	int ra, ex, i;
+	int ra, i;
 	char path[256], *p, **nrest;
 
-	if (ifname == NULL) {
-		if (ipv6_checked)
-			return 1;
-		ipv6_checked = 1;
+	if (ifname == NULL)
 		ifname = "all";
-		ex = 1;
-	} else
-		ex = 0;
 
 	snprintf(path, sizeof(path), "%s/%s/autoconf", prefix, ifname);
 	i = check_proc_int(path);
@@ -213,27 +184,24 @@ check_ipv6(const char *ifname, int own)
 			syslog(LOG_ERR, "write_path: %s: %m", path);
 			return ra;
 		}
-		for (i = 0; i < nrestore; i++)
-			if (strcmp(restore[i], ifname) == 0)
+		for (i = 0; i < ctx->ra_restore_len; i++)
+			if (strcmp(ctx->ra_restore[i], ifname) == 0)
 				break;
-		if (i == nrestore) {
+		if (i == ctx->ra_restore_len) {
 			p = strdup(ifname);
 			if (p == NULL) {
 				syslog(LOG_ERR, "%s: %m", __func__);
 				return ra;
 			}
-			nrest = realloc(restore,
-			    (nrestore + 1) * sizeof(char *));
+			nrest = realloc(ctx->ra_restore,
+			    (ctx->ra_restore_len + 1) * sizeof(char *));
 			if (nrest == NULL) {
 				syslog(LOG_ERR, "%s: %m", __func__);
 				return ra;
 			}
-			restore = nrest;
-			restore[nrestore++] = p;
-
+			ctx->ra_restore = nrest;
+			ctx->ra_restore[ctx->ra_restore_len++] = p;
 		}
-		if (ex)
-			atexit(restore_kernel_ra);
 	}
 
 	return ra;
